/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#if !defined(HARK_BASE_VECTOR_H_INCLUDED)
#define HARK_BASE_VECTOR_H_INCLUDED 1

#include "common.h"

#include "Object.h"
#include <vector>
#include "ObjectParser.h"
#include "ObjectRef.h"
#include <string>
#include <complex>

namespace HARK
{
    template<class T>
    class Vector : public Object, public std::vector<T>
    {
    public:
	using basicType = T;

	Vector() :
	    Object(),
	    std::vector<T> ()
	{}

	Vector(const std::vector<T> &v) :
	    Object(),
	    std::vector<T> (v)
	{
	}

	Vector(const Vector<T> &v) :
	    Object(),
	    std::vector<T> (v)
	{
	}

	explicit Vector(size_t n, const T &x = T()) :
	    Object(),
	    std::vector<T> (n, x)
	{}

	~Vector()
	{
	}

	virtual size_t vsize() const
	{
	    return this->size();
	}

	virtual bool vempty() const
	{
	    return this->empty();
	}

	void prettyPrint(std::ostream &out=std::cout) const;

	void printOn(std::ostream &out) const;

	void readFrom(std::istream &in=std::cin);

	virtual void destroy();

	static Vector<T> *alloc(size_t size);

	static std::string GetClassName()
	{
	    std::string name = ObjectGetClassName<Vector<T> >();
	    if (name == "unknown") {
		return std::string("Vector");
	    }
	    else {
		return name;
	    }
	}

	std::string getClassName()
	{
	    return GetClassName();
	}

	ObjectRef range(size_t startInd, size_t endInd)
	{
	    Vector<T> *v = Vector<T>::alloc(endInd-startInd+1);
	    if (endInd >= v->size() || startInd < 0) {
		throw new GeneralException("Index out of range in BaseVector::range()", __FILE__, __LINE__);
	    }
	    for (size_t i=startInd; i<=endInd; i++) {
		(*v)[i-startInd]=(*this)[i];
	    }
	    return ObjectRef(v);
	}

	virtual ObjectRef clone();
    };

    template <class T>
    inline ObjectRef Vector<T>::clone()
    {
	Vector<T> *cpy = Vector<T>::alloc(this->size());

	for (unsigned int i = 0; i < this->size(); i++) {
	    (*cpy)[i] = (*this)[i];
	}
	return ObjectRef(cpy);
    }

    template <>
    inline ObjectRef Vector<ObjectRef>::clone()
    {

	Vector<ObjectRef> *cpy = new Vector<ObjectRef>(this->size());

	for (unsigned int i = 0; i < this->size(); i++) {
	    (*cpy)[i] = (*this)[i]->clone();
	}

	return ObjectRef(cpy);
    }

    template <class T>
    inline void _vector_printOn(const Vector<T> &v, std::ostream &out)
    {
	out << "<" << v.className();
	for (size_t i=0; i < v.size(); i++) {
	    out << " " << v[i];
	}
	out << " > ";
    }

    template <>
    inline void _vector_printOn(const Vector<std::string> &v, std::ostream &out)
    {
	out << "<Vector<string>";
	for (unsigned int n=0; n < v.size(); n++) {
	    out << " ";
	    const std::string &str = v[n];
	    for (unsigned int i=0; i<str.size(); i++) {
		if (str[i] == '>') {
		    out.put('\\');
		    out.put('>');
		} else if (str[i] == ' ') {
		    out.put('\\');
		    out.put(' ');
		} else if (str[i] == '\\') {
		    out.put('\\');
		    out.put('\\');
		} else {
		    out.put(str[i]);
		}
	    }
	}
	out << "> ";
    }

    template <class T>
    inline void _vector_printOn(const Vector<T*> &v, std::ostream &out)
    {
	out << "<" << v.className();
	for (size_t i=0; i < v.size(); i++) {
	    out << " " << *(v[i]);
	}
	out << " > ";
    }

    template <class T>
    void Vector<T>::printOn(std::ostream &out) const
    {
	_vector_printOn(*this, out);
    }

    template <class T>
    void Vector<T>::prettyPrint(std::ostream &out) const
    {
	_vector_printOn(*this,out);
    }

    template <class T>
    inline void _vector_readFrom(Vector<T> &v, std::istream &in)
    {
	bool a=false;
	v.resize(0);
	while (1) {
	    char ch=' ';
	    while (ch == ' ') {
		in >> ch;
		if (ch == '>') {
		    return;
		} else if (ch != ' ') {
		    in.putback(ch);
		}
		if (in.fail()) {
		    //throw new GeneralException("Error reading Vector: '>' expected", __FILE__, __LINE__);
		    a=true;
		    break;
		}
	    }
	    T tmp;
	    in >> tmp;
	    if (in.fail()) {
		throw new GeneralException("Error reading Vector", __FILE__, __LINE__);
	    }
	    v.push_back(tmp);
	}
	if (a) {
	    throw new GeneralException("Error reading Vector: '>' expected", __FILE__, __LINE__);
	}
    }

    template <>
    inline void _vector_readFrom(Vector<std::string> &v, std::istream &in)
    {
	bool done=false;
	while (1) {
	    std::string tmp;
	    int i=0;
	    while(1) {
		char ch;
		in.get(ch);
		if (in.eof() || in.fail()) {
		    throw new GeneralException("Error reading String: '>' or '}' expected", __FILE__, __LINE__);
		}
		if (ch == '\\') {
		    in.get(ch);
		    tmp += ch;
		} else if (ch == ' ') {
		    if (i) {
			break;
		    } else {
			continue;
		    }
		} else if (ch == '>') {
		    done=true;
		    break;
		} else if (ch == '}') {
		    break;
		} else {
		    tmp += ch;
		}
		i++;
	    }

	    if (tmp != "") {
		v.push_back(tmp);
	    }
	    if (done) {
		break;
	    }
	}
    }

    template <class T>
    inline void _vector_readFrom(Vector<T*> &v, std::istream &in)
    {
	v.resize(0);
	while (1) {
	    char ch=' ';
	    while (ch == ' ') {
		in >> ch;
		if (ch == '>') {
		    return;
		} else if (ch != ' ') {
		    in.putback(ch);
		}
		if (in.fail()) {
		    throw new GeneralException("Error reading Vector: '>' expected", __FILE__, __LINE__);
		}
	    }
	    T *tmp = new T;
	    in >> *tmp;
	    if (in.fail()) {
		throw new GeneralException("Error reading Vector", __FILE__, __LINE__);
	    }
	    v.push_back(tmp);
	}
    }

    template <class T>
    inline void Vector<T>::readFrom(std::istream &in)
    {
	_vector_readFrom(*this, in);
    }

//    template<class T, int I>
//    struct VecMethod {
//    };

    template <class T>
    inline void Vector<T>::destroy()
    {
	delete this;
    }

    template <class T>
    inline Vector<T> *Vector<T>::alloc(size_t size)
    {
	return new Vector<T> (size);
    }

    inline bool isValidVectorType (std::istream &in, std::string type, const std::string &className)
    {
	if (type != "Vector" && type != className) {
	    return false;
	}
	return true;
    }

    template<class T>
    std::istream &operator >> (std::istream &in, Vector<T> &vec)
    {
	char ch;
	in >> ch;

	std::string expected = ObjectGetClassName<Vector<T> >();

	if (ch == '<') {
	    std::string type;
	    in >> type;
	    if (!isValidVectorType(in, type, expected)) {
		throw new ParsingException ("Parser expected type " + expected + " and got " + type);
	    }
	    vec.readFrom(in);
	} else if (ch == '{') {
	    std::string type;
	    in >> type;
	    if (!isValidVectorType(in, type, expected)) {
		throw new ParsingException ("Parser expected type " + expected + " and got " + type);
	    }
	    char dummy;
	    do {
		in >> dummy;
	    } while(dummy != '|');
	} else {
	    throw new ParsingException ("Parser expected < or { while parsing type " + expected);
	}
	return in;
    }

    template<class T>
    inline Vector<T> &operator+= (Vector<T> &v1, const Vector<T> &v2)
    {
	if (v1.size() != v2.size()) {
	    throw new GeneralException("Vector size mismatch", __FILE__, __LINE__);
	}
	for (size_t i=0; i<v1.size(); i++) {
	    v1[i] += v2[i];
	}
	return v1;
    }

    template<class T>
    inline Vector<T> &operator-= (Vector<T> &v1, const Vector<T> &v2)
    {
	if (v1.size() != v2.size()) {
	    throw new GeneralException("Vector size mismatch", __FILE__, __LINE__);
	}
	for (size_t i=0; i<v1.size(); i++) {
	    v1[i] -= v2[i];
	}
	return v1;
    }

    template<class T>
    inline Vector<T> &operator*= (Vector<T> &v1, const Vector<T> &v2)
    {
	if (v1.size() != v2.size()) {
	    throw new GeneralException("Vector size mismatch", __FILE__, __LINE__);
	}
	for (size_t i=0; i<v1.size(); i++) {
	    v1[i] *= v2[i];
	}
	return v1;
    }

    template<class T>
    inline Vector<T> &operator/= (Vector<T> &v1, const Vector<T> &v2)
    {
	if (v1.size() != v2.size()) {
	    throw new GeneralException("Vector size mismatch", __FILE__, __LINE__);
	}
	for (size_t i=0; i<v1.size(); i++) {
	    v1[i] /= v2[i];
	}
	return v1;
    }


    template<class T>
    inline Vector<T> operator+ (const Vector<T> &v1, const Vector<T> &v2)
    {
	Vector<T> v(v1);
	v += v2;
	return v;
    }

    template<class T>
    inline Vector<T> operator- (const Vector<T> &v1, const Vector<T> &v2)
    {
	Vector<T> v(v1);
	v -= v2;
	return v;
    }

    template<class T>
    inline Vector<T> operator* (const Vector<T> &v1, const Vector<T> &v2)
    {
	Vector<T> v(v1);
	v *= v2;
	return v;
    }

    template<class T>
    inline Vector<T> operator/ (const Vector<T> &v1, const Vector<T> &v2)
    {
	Vector<T> v(v1);
	v /= v2;
	return v;
    }

    template<class T>
    inline Vector<T> operator- (Vector<T> &v1)
    {
	Vector<T> v(v1.size());
	for (size_t i=0; i<v1.size(); i++) {
	    v[i] = -v1[i];
	}
	return v;
    }

    template<class T>
    inline Vector<T> &operator+= (Vector<T> &v1, T scal)
    {
	for (size_t i=0; i<v1.size(); i++) {
	    v1[i] += scal;
	}
	return v1;
    }

    template<class T>
    inline Vector<T> &operator-= (Vector<T> &v1, T scal)
    {
	for (size_t i=0; i<v1.size(); i++) {
	    v1[i] -= scal;
	}
	return v1;
    }

    template<class T>
    inline Vector<T> &operator*= (Vector<T> &v1, T scal)
    {
	for (size_t i=0; i<v1.size(); i++) {
	    v1[i] *= scal;
	}
	return v1;
    }

    template<class T>
    inline Vector<T> &operator/= (Vector<T> &v1, T scal)
    {
	for (size_t i=0; i<v1.size(); i++) {
	    v1[i] /= scal;
	}
	return v1;
    }

    template<class T>
    inline Vector<T> operator+ (const Vector<T> &v1, T scal)
    {
	Vector<T> v(v1);
	v += scal;
	return v;
    }

    template<class T>
    inline Vector<T> operator- (const Vector<T> &v1, T scal)
    {
	Vector<T> v(v1);
	v -= scal;
	return v;
    }

    template<class T>
    inline Vector<T> operator* (const Vector<T> &v1, T scal)
    {
	Vector<T> v(v1);
	v *= scal;
	return v;
    }

    template<class T>
    inline Vector<T> operator/ (const Vector<T> &v1, T scal)
    {
	Vector<T> v(v1);
	v /= scal;
	return v;
    }

    template<>
    inline void Vector<float>::prettyPrint(std::ostream &out) const
    {
	for (unsigned int i = 0; i < size(); i++) {
	    out<<(*this)[i]<<" ";
	}
	out<<std::endl;
    }

    template<>
    inline void Vector<double>::prettyPrint(std::ostream &out) const
    {
	for (unsigned int i = 0; i < size(); i++) {
	    out<<(*this)[i]<<" ";
	}
	out<<std::endl;
    }

    template<>
    inline void Vector<int>::prettyPrint(std::ostream &out) const
    {
	for (unsigned int i = 0; i < size(); i++) {
	    out<<(*this)[i]<<" ";
	}
	out<<std::endl;
    }

    template<>
    inline void Vector<ObjectRef>::prettyPrint(std::ostream &out) const
    {
	for (unsigned int i = 0; i < size(); i++) {
	    (*this)[i]->prettyPrint(out);
	    out<<std::endl;
	}
    }

    template<>
    inline void Vector<std::complex<float> >::prettyPrint(std::ostream &out) const
    {
	for (unsigned int i = 0; i < size(); i++) {
	    out<<(*this)[i]<<" ";
	}
	out<<std::endl;
    }

    template<>
    inline void Vector<std::complex<double> >::prettyPrint(std::ostream &out) const
    {
	for (unsigned int i = 0; i < size(); i++) {
	    out<<(*this)[i]<<" ";
	}
	out<<std::endl;
    }

    template<>
    inline void Vector<std::string>::prettyPrint(std::ostream &out) const
    {
	for (unsigned int i = 0; i < size(); i++) {
	    out<<(*this)[i]<<std::endl;
	}
    }

    template<>
    inline void Vector<String>::prettyPrint(std::ostream &out) const
    {
	for (unsigned int i = 0; i < size(); i++) {
	    out<<(*this)[i]<<std::endl;
	}
    }
}

#endif

/* end of file */
