/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#if !defined(HARK_BASE_NET_TYPES_H_INCLUDED)
#define HARK_BASE_NET_TYPES_H_INCLUDED 1

#include "common.h"

#include "Object.h"
#include "ObjectParser.h"

#include <complex>
#include <iostream>
#include <string>

namespace HARK
{
    template<typename T>
    class GenericType : public Object
    {
    protected:
	T value;

    public:
	using basicType = T;

	GenericType();
	GenericType(T val);
	GenericType(const GenericType<T> &src);

	virtual ~GenericType();

	T &val();
	
	virtual void printOn(std::ostream &out = std::cout) const;
    };

    template<typename T>
    class PrintableGenericType : public GenericType<T>
    {
    public:
	PrintableGenericType();
	PrintableGenericType(T val);
	PrintableGenericType(const PrintableGenericType<T> &src);

	T &val();

	void printOn(std::ostream &out) const;

	virtual void prettyPrint(std::ostream &out = std::cout) const;

	void readFrom(std::istream &in);

	void serialize(std::ostream &out) const;

	void unserialize(std::istream &in);
    };

    template<typename T>
    class NetCType : public PrintableGenericType<T>
    {
    public:
	NetCType();
	NetCType(T val);

	operator T ();
	NetCType<T> &operator = (const NetCType<T> &type);
	NetCType<T> &operator = (T val);
	int operator == (const NetCType<T> &type) const;
	int operator == (T val) const;
	int operator != (const NetCType<T> &type) const;
	int operator != (T val) const;

	static NetCType<T> *alloc();
	static NetCType<T> *alloc(const T &obj);
	void destroy();
    };

    using Char = NetCType<char>;
    using Int = NetCType<int>;
    using Float = NetCType<float>;
    using Double = NetCType<double>;
    using Bool = NetCType<bool>;
    extern ObjectRef TrueObject;
    extern ObjectRef FalseObject;

    class FILEPTR : public GenericType<FILE *>
    {
    public:
	FILEPTR(FILE *file);
	FILEPTR(const std::string &filename, const std::string &mode);
	~FILEPTR();
    };

#ifndef WIN32

    class FILEDES : public GenericType<int>
    {
    public:
	FILEDES(int fd);
	FILEDES(const std::string &filename, int mode);
	~FILEDES();
    };

#endif /*ifdef WIN32*/

    class String : public std::string, public Object
    {
    public:
	using basicType = std::string;
	
	String();
	String(const char *str);
	String(const std::string &str);

	const std::string &val() const;

	void printOn(std::ostream &out) const;
	void readFrom(std::istream &in);
	void prettyPrint(std::ostream &out) const;
    };

    template<typename T>
    GenericType<T>::GenericType()
    {
    }
    
    template<typename T>
    GenericType<T>::GenericType(T val) :
	value(val)
    {
    }

    template<typename T>
    GenericType<T>::GenericType(const GenericType<T> &src) :
	value(src.value)
    {
    }

    template<typename T>
    GenericType<T>::~GenericType()
    {
    }

    template<typename T>
    T &
    GenericType<T>::val()
    {
	return value;
    }

    template<typename T>
    void
    GenericType<T>::printOn(std::ostream &out) const
    {
	out << "<Generic Type: " << typeid(T).name() << ">" << std::endl;
    }

    template<typename T>
    PrintableGenericType<T>::PrintableGenericType()
    {
    }

    template<typename T>
    PrintableGenericType<T>::PrintableGenericType(T val) :
	GenericType<T>(val)
    {
    }

    template<typename T>
    PrintableGenericType<T>::PrintableGenericType(const PrintableGenericType<T> &src) :
	GenericType<T>(src)
    {
    }

    template<typename T>
    T &
    PrintableGenericType<T>::val()
    {
	return GenericType<T>::value;
    }

    template<typename T>
    void
    PrintableGenericType<T>::printOn(std::ostream &out) const
    {
	out << "<" << this->className() << " " << this->value << " >";
    }

    template<typename T>
    void
    PrintableGenericType<T>::prettyPrint(std::ostream &out) const
    {
	out << this->value << " ";
    }

    template<typename T>
    void
    PrintableGenericType<T>::readFrom(std::istream &in)
    {
	in >> this->value;
	char ch;
	in >> ch;
	if (ch != '>') {
	    throw new GeneralException("Error reading String: '>' expected", __FILE__, __LINE__);
	}
    }

    template<typename T>
    void
    PrintableGenericType<T>::serialize(std::ostream &out) const
    {
	out << "{" << this->className() << " |" << this->value << " }";
    }

    template<typename T>
    void
    PrintableGenericType<T>::unserialize(std::istream &in)
    {
	in >> this->value;
	char ch;
	in >> ch;
	if (ch != '}') {
	    throw new GeneralException("Error reading String: '}' expected", __FILE__, __LINE__);
	}
    }

    template<typename T>
    NetCType<T>::NetCType() :
	PrintableGenericType<T>((T)0)
    {
    }

    template<typename T>
    NetCType<T>::NetCType(T val) :
	PrintableGenericType<T>(val)
    {
    }

    template<typename T>
    NetCType<T>::operator T()
    {
	return GenericType<T>::value;
    }

    template<typename T>
    NetCType<T> &
    NetCType<T>::operator = (const NetCType<T> &type)
    {
	GenericType<T>::value = type.value;
	return *this;
    }

    template<typename T>
    NetCType<T> &
    NetCType<T>::operator = (T val)
    {
	GenericType<T>::value = val;
	return *this;
    }

    template<typename T>
    int
    NetCType<T>::operator == (const NetCType<T> &type) const
    {
	return (type.value == GenericType<T>::value);
    }

    template<typename T>
    int
    NetCType<T>::operator == (T val) const
    {
	return (GenericType<T>::value == val);
    }

    template<typename T>
    int
    NetCType<T>::operator != (const NetCType<T> &type) const
    {
	return (type.value != GenericType<T>::value);
    }

    template<typename T>
    int
    NetCType<T>::operator != (T val) const
    {
	return (GenericType<T>::value != val);
    }

    template<typename T>
    NetCType<T> *
    NetCType<T>::alloc()
    {
	return new NetCType<T>();
    }

    template<typename T>
    NetCType<T> *
    NetCType<T>::alloc(const T &obj)
    {
	NetCType<T> *ret = new NetCType<T>();
	ret->value = obj;
	return ret;
    }

    template<typename T>
    void
    NetCType<T>::destroy()
    {
	delete this;
    }

    using Char = NetCType<char>;
    using Int = NetCType<int>;
    using Float = NetCType<float>;
    using Double = NetCType<double>;
    using Bool = NetCType<bool>;

    std::istream &operator >> (std::istream &in, String &str);

    template <class T>
    class Complex : public std::complex<T>, public Object
    {
    public:
	using basicType = std::complex<T>;

	Complex();
	Complex(const std::complex<T> &val);
	Complex(const Complex<T> &val);
	Complex(const NetCType<std::complex<T> > &obj);

	void printOn(std::ostream &out) const;
	void readFrom(std::istream &in);
	void prettyPrint(std::ostream &out) const;
	std::complex<T>& val() const;

	static Complex<T> *alloc();
	static Complex<T> *alloc(const Complex<T> &obj);

	void destroy();

	virtual ObjectRef clone();
	
    };

    template<typename T>
    Complex<T>::Complex() :
	std::complex<T>()
    {
    }

    template<typename T>
    Complex<T>::Complex(const std::complex<T> &val) :
	std::complex<T>(val)
    {
    }
    
    template<typename T>
    Complex<T>::Complex(const Complex<T> &val) :
	std::complex<T>(val)
    {
    }
    
    template<typename T>
    Complex<T>::Complex(const NetCType<std::complex<T> > &obj) :
	std::complex<T>(obj.val())
    {
    }

    template<typename T>
    void
    Complex<T>::printOn(std::ostream &out) const
    {
	out << "<"<<className()<<" ";
	out << *((std::complex<T>*) this);
	out << ">"<<std::endl;
    }

    template<typename T>
    void
    Complex<T>::readFrom(std::istream &in)
    {
	std::complex<T> value;
	in >> value;
	*this = value;
	char ch;
	in >> ch;
	if (ch != '>') {
	    throw new GeneralException(std::string("Error reading ") + className() + std::string(" > expected"), __FILE__, __LINE__);
	}
    }

    template<typename T>
    void
    Complex<T>::prettyPrint(std::ostream &out) const
    {
	out << *((std::complex<T>*) this);
    }

    template<typename T>
    std::complex<T>&
    Complex<T>::val() const
    {
	return *((std::complex<T>*) this);
    }

    template<typename T>
    Complex<T> *
    Complex<T>::alloc()
    {
	return new Complex<T>();
    }

    template<typename T>
    Complex<T> *
    Complex<T>::alloc(const Complex<T> &obj)
    {
	Complex<T> *ret = new Complex<T>();
	*ret = obj;
	return ret;
    }

    template<typename T>
    void
    Complex<T>::destroy()
    {
	delete this;
    }

    template<typename T>
    ObjectRef
    Complex<T>::clone()
    {
	return ObjectRef(Complex<T>::alloc(*(std::complex<T>*) this));
    }

    template <class T>
    std::istream &
    operator >> (std::istream &in, Complex<T> &value)
    {
	char ch;
	in >> ch;

	std::string expected = ObjectGetClassName<Complex<T> >();

	if (ch == '<') {
	    std::string type;
	    in >> type;

	    if (expected != type) {
		throw new ParsingException("Parser expected type " + expected + " and got " + type);
	    }

	    value.readFrom(in);
	    
	} else if (ch == '{') {
	    std::string type;
	    in >> type;

	    if (expected != type) {
		throw new ParsingException("Parser expected type " + expected + " and got " + type);
	    }

	    char dummy;
	    do {
		in >> dummy;
	    } while(dummy != '|');

	    value.unserialize(in);

	} else {
	    throw new ParsingException ("Parser expected < or { while parsing type " + expected);
	}

	return in;
    }
}

#endif

/* end of file */
