/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "Buffer.h"

namespace HARK
{
    Buffer::Buffer(int bLength) :
	data        (bLength),
	flags       (bLength, 0),
	bufferLength(bLength)
    {
	bufferPos  = -1;
	currentPos = -1;
    }

    Buffer::Buffer(const Buffer &)
    {
	throw new BufferException(NULL,"use an ObjectRef instead",0);
    }

    ObjectRef &
    Buffer::get(int ind) const
    {
	if (ind < 0 || ind <= currentPos-bufferLength || ind > currentPos) {
	    throw new BufferException (this, "trying to read non-existing element",ind);
	}
	int tmp = bufferPos+ind-currentPos;
	if (tmp < 0) {
	    tmp += bufferLength;
	}
	if (flags[tmp]) {
	    return data[tmp];
	} else {
	    throw new BufferException (this, "trying to read not initialized element",ind);
	}
    }

    ObjectRef &
    Buffer::operator[] (int ind)
    {
	if (ind < 0 || ind <= currentPos-bufferLength) {
	    throw new BufferException (this, "trying to write to non-existing element",ind);
	}
	if (ind > currentPos) {
	    int diff = ind-currentPos;
	    while (diff--) {
		bufferPos++;
		if (bufferPos == bufferLength) {
		    bufferPos=0;
		}
		flags[bufferPos] = 0;
	    }
	    currentPos = ind;
	    flags[bufferPos] = 1;
	    return data[bufferPos];
	}

	int tmp = bufferPos+ind-currentPos;
	if (tmp < 0) {
	    tmp += bufferLength;
	}
	flags[tmp] = 1;
	return data[tmp];
    }

    int
    Buffer::isValid(int ind) const
    {
	if (ind > currentPos || ind <= currentPos-bufferLength) {
	    return false;
	}
	int tmp = bufferPos+ind-currentPos;
	if (tmp < 0) {
	    tmp += bufferLength;
	}
	return flags[tmp];
    }

    void
    Buffer::printOn(std::ostream &out) const
    {
	int i;
	out << "<Buffer" << std::endl;
	for (i=currentPos-bufferLength+1; i<=currentPos; i++) {
	    if (i>=0) {
		out << "< " << i << " ";
		if (isValid(i)) {
		    out << get(i);
		} else {
		    out << "nil";
		}
	    }
	}
	out << " >" << std::endl;
    }

    int
    Buffer::getCurrentPos()
    {
	return currentPos;
    }

    BufferException::BufferException(const Buffer *_buffer, std::string _message, int _element) :
	buffer (_buffer),
	message(_message),
	element(_element)
    {
    }

    void
    BufferException::print(std::ostream &out) const
    {
	out<< typeid(buffer).name() << " error: "<< message << ".\nElement " << element << std::endl;
	out << "Buffer is: \n";
	out << *buffer;
    }
}

/* end of file */
