/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "InputStream.h"

#include <fcntl.h>
#include <unistd.h>
#include "Stream.h"

#include <iostream>
#include <fstream>

#ifdef WIN32

#include <windows.h>
#define sleep(x) Sleep((x)*1000)

#endif

namespace HARK
{
    DECLARE_NODE(InputStream);
    /*Node
     *
     * @name InputStream
     * @category IO
     * @description Creates a read-only stream from a filename
     *
     * @input_name INPUT
     * @input_description The file name
     * @input_type string
     *
     * @output_name OUTPUT
     * @output_description The Stream
     * @output_type Stream
     *
     * @parameter_name TYPE
     * @parameter_type String
     * @parameter_description Type of stream: stream, fd, or FILE (default stream)
     *
     * @parameter_name RETRY
     * @parameter_type int
     * @parameter_description If set to N, InputStream will retry N times on open fail
     *
     END*/

    InputStream::InputStream()
    {
        throw new GeneralException("InputStream copy constructor should not be called",__FILE__,__LINE__);
    }

    InputStream::InputStream(const std::string &nodeName, const ParameterSet &params) :
        BufferedNode(nodeName, params)
    {
        this->_outputID = this->addOutput("OUTPUT");
        this->_inputID  = this->addInput("INPUT");

        if (parameters.exist("TYPE")) {
            const String &strType = object_cast<String> (parameters.get("TYPE"));
            if (strType == "stream") {
                this->_type = cpp;
            } else if (strType == "FILE") {
                this->_type = fptr;
            }
#ifndef WIN32
            else if (strType == "fd") {
                this->_type = fd;
            }
#endif
            else {
                throw new NodeException(NULL, "Bad stream type: " + strType, __FILE__, __LINE__);
            }

        } else {
            this->_type = cpp;
        }

        if (parameters.exist("RETRY")) {
            this->_retry = dereference_cast<int> (parameters.get("RETRY"));
        } else {
            this->_retry = 0;
        }
    }

    void
    InputStream::calculate(int output_id, int count, Buffer &out)
    {
        ObjectRef inputValue = this->getInput(this->_inputID, count);
        const String &fileName = object_cast<String> (inputValue);

        ObjectRef openedFile;
        switch (this->_type) {
        case cpp: {

            std::ifstream *file = new std::ifstream(fileName.c_str(), std::ios::binary);

            if (this->_retry) {
                int retryLeft=this->_retry;
                int retryTime=5;
                while (file->fail() && retryLeft--) {
		    std::cerr << "deleting\n";
                    delete file;
                    sleep (retryTime);
                    retryTime *= 2;
                    file = new std::ifstream(fileName.c_str(), std::ios::binary);
                    std::cerr << "retrying\n";
                    std::cerr << file->fail() << " " << retryLeft << std::endl;
                }
            }
            if (file->fail()) {
                delete file;
                throw new NodeException(this, "InputStream: cannot open file: " + fileName, __FILE__, __LINE__);
            }
            openedFile = ObjectRef (new IStream(file));
        }
        break;

        case fptr: {
            FILE *file = fopen (fileName.c_str(), "rb");
            if (!file) {
                throw new NodeException(this, "InputStream: cannot open file: " + fileName, __FILE__, __LINE__);
            }
            openedFile = ObjectRef (new FILEPTR(file));
        }
        break;

#ifndef WIN32
        case fd: {
            int file = open (fileName.c_str(), O_RDONLY);
            if (file == -1) {
                throw new NodeException(this, "InputStream: cannot open file: " + fileName, __FILE__, __LINE__);
            }
            openedFile = ObjectRef (new FILEDES(file));
        }
        break;
#endif

        }

        out[count] = openedFile;
    }
}

/* end of file */
