/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "Sleep.h"

namespace HARK
{
    DECLARE_NODE(Sleep);
    /*Node
     *
     * @name Sleep
     * @category Flow
     * @description Sleep a certain amount of time.
     *
     * @output_name VALUE
     * @output_description Always return TRUE.
     *
     * @parameter_name SECONDS
     * @parameter_description Sleep x seconds.
     * @parameter_type float
     *
     END*/

    static bool _sleep_is_stop = false;

    static void signal_handler(int s)
    {
        if(s == SIGINT){ _sleep_is_stop = true; }
    }

    Sleep::Sleep(const std::string &nodeName, const ParameterSet &params) :
        Node(nodeName, params)
    {
        float sec = dereference_cast<float>(parameters.get("SECONDS"));
        this->_is_long = (bool) ((int) (sec - 2.5f) > 0);
        this->_usec = sec * 1000000;
        this->_start = std::chrono::high_resolution_clock::now();
        this->_outputID = this->addOutput("VALUE");
        std::signal(SIGINT,  signal_handler);
    }

    ObjectRef Sleep::getOutput(int output_id, int count)
    {
        if(_sleep_is_stop){
            std::cout << std::endl << "User stop" << std::endl;
            throw new NodeException(this, "Sleep: Interrupted by the keyboard.", __FILE__, __LINE__);
        }
        {
            auto t = this->_start + std::chrono::microseconds(this->_usec) * (count + 1);
            if(_is_long){
                do{
                    std::this_thread::sleep_for(std::chrono::seconds(1));
                    if(_sleep_is_stop){
                        std::cout << std::endl << "User stop" << std::endl;
                        throw new NodeException(this, "Sleep: Interrupted by the keyboard.", __FILE__, __LINE__);
                    }
                } while(t - std::chrono::high_resolution_clock::now() > std::chrono::seconds(2));
            }
            std::this_thread::sleep_for(t - std::chrono::high_resolution_clock::now());
        }
        if(output_id == this->_outputID){
            return TrueObject;
        }
        else{
            throw new NodeException(this, "Sleep: Unknown output id", __FILE__, __LINE__);
        }
    }
}

/* end of file */
