/*
 * Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 * 
 * Licensed under the HARK license, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * https://www.hark.jp/notice/HARK_License_Agreement.pdf
 */

#include "operators.h"

#include "Vector.h"
#include "Matrix.h"

#include <typeinfo>

namespace HARK
{
    using B  = Bool;
    using SI = Int;
    using SF = Float;
    using SD = Double;
    using SG = Complex<float>;
    using SE = Complex<double>;
    using VI = Vector<int>;
    using VF = Vector<float>;
    using VD = Vector<double>;
    using VG = Vector<std::complex<float>>;
    using VE = Vector<std::complex<double>>;
    using MI = Matrix<int>;
    using MF = Matrix<float>;
    using MD = Matrix<double>;
    using MG = Matrix<std::complex<float>>;
    using ME = Matrix<std::complex<double>>;

    template<typename X, typename Y, typename Z>
    struct NullOperator
    {
	static Z Operator(X x, Y y) {
	    return 0;
	}
    };

    template<typename X, typename Y, typename Z>
    struct EqualOperator
    {
	static Z Operator(X x, Y y) {
	    return (x == y);
	}
    };

    template<typename X, typename Y, typename Z>
    struct AddOperator
    {
	static Z Operator(X x, Y y) {
	    return (x + y);
	}
    };

    template<typename X, typename Y, typename Z>
    struct SubOperator
    {
	static Z Operator(X x, Y y) {
	    return (x - y);
	}
    };

    template<typename X, typename Y, typename Z>
    struct MulOperator
    {
	static Z Operator(X x, Y y) {
	    return (x * y);
	}
    };

    template<typename X, typename Y, typename Z>
    struct DivOperator
    {
	static Z Operator(X x, Y y) {
	    return (x / y);
	}
    };

    template<typename X, typename Y, typename Z>
    struct MinOperator
    {
	static Z Operator(X x, Y y) {
	    return (x < y) ? x : y;
	}
    };

    template<typename X, typename Y, typename Z>
    struct MaxOperator
    {
	static Z Operator(X x, Y y) {
	    return (x > y) ? x : y;
	}
    };
    
    template<typename X, typename Y, typename Z>
    struct LessOperator
    {
	static Z Operator(X x, Y y) {
	    //std::cout << "LessOperator: x=" << x << " [" << typeid(x).name() << "] , y=" << y << " [" << typeid(y).name() << "], r=" << (x < y) << std::endl;
	    return (x < y);
	}
    };

    template<typename X, typename Y, typename Z>
    struct GreaterOperator
    {
	static Z Operator(X x, Y y) {
	    return (x > y);
	}
    };
    
    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct ScalarScalarFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;
	    RCPtr<W> vz(W::alloc(static_cast<typename W::basicType>(OP<typename Z::basicType, typename Z::basicType, typename W::basicType>::Operator(static_cast<typename Z::basicType>(vx->val()), static_cast<typename Z::basicType>(vy->val())))));
	return vz;
	}
    };

    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct VectorVectorFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;
	    if(vx->size() != vy->size()) {
		throw new GeneralException("Vector size mismatch ",__FILE__,__LINE__);
	    }

	    RCPtr<W> vz(W::alloc(vx->size()));

	    for (int i = 0; i < vz->size(); i++) {    
		(*vz)[i] = static_cast<typename W::basicType>(OP<typename Z::basicType, typename Z::basicType, typename W::basicType>::Operator(static_cast<typename Z::basicType>((*vx)[i]), static_cast<typename Z::basicType>((*vy)[i])));
	    }

	    return vz;
	}
    };

    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct VectorScalarFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;

	    RCPtr<W> vz(W::alloc(vx->size()));

	    for (int i = 0; i < vz->size(); i++) {    
		(*vz)[i] = static_cast<typename W::basicType>(OP<typename Z::basicType, typename Z::basicType, typename W::basicType>::Operator(static_cast<typename Z::basicType>((*vx)[i]), static_cast<typename Z::basicType>(vy->val())));
	    }

	    return vz;
	}
    };

    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct ScalarVectorFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;

	    RCPtr<W> vz(W::alloc(vy->size()));

	    for (int i = 0; i < vz->size(); i++) {    
		(*vz)[i] = static_cast<typename W::basicType>(OP<typename Z::basicType, typename Z::basicType, typename W::basicType>::Operator(static_cast<typename Z::basicType>(vx->val()), static_cast<typename Z::basicType>((*vy)[i])));
	    }

	    return vz;
	}
    };

    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct MatrixMatrixFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;
	    if(vx->nrows() != vy->nrows() || vx->ncols() != vy->ncols()) {
		throw new GeneralException("Matrix size mismatch ",__FILE__,__LINE__);
	    }

	    RCPtr<W> vz(new W(vx->nrows(), vx->ncols()));

	    for (int i = 0; i < vz->nrows(); i++) {    
		for (int j = 0; j < vz->ncols(); j++) {
		    (*vz)(i, j) = static_cast<typename W::basicType>(OP<typename Z::basicType, typename Z::basicType, typename W::basicType>::Operator(static_cast<typename Z::basicType>((*vx)(i, j)), static_cast<typename Z::basicType>((*vy)(i, j))));
		}
	    }

	    return vz;
	}
    };

    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct MatrixScalarFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;

	    RCPtr<W> vz(new W(vx->nrows(), vx->ncols()));

	    for (int i = 0; i < vz->nrows(); i++) {    
		for (int j = 0; j < vz->ncols(); j++) {
		    (*vz)(i, j) = static_cast<typename W::basicType>(OP<typename Z::basicType, typename Z::basicType, typename W::basicType>::Operator(static_cast<typename Z::basicType>((*vx)(i, j)), static_cast<typename Z::basicType>(vy->val())));
		}
	    }

	    return vz;
	}
    };

    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct ScalarMatrixFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;

	    RCPtr<W> vz(new W(vy->nrows(), vy->ncols()));

	    for (int i = 0; i < vz->nrows(); i++) {    
		for (int j = 0; j < vz->ncols(); j++) {
		    (*vz)(i, j) = static_cast<typename W::basicType>(OP<typename Z::basicType, typename Z::basicType, typename W::basicType>::Operator(static_cast<typename Z::basicType>(vx->val()), static_cast<typename Z::basicType>((*vy)(i, j))));
		}
	    }

	    return vz;
	}
    };

    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct ConcatenateScalarScalarFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;
	    RCPtr<W> vz(W::alloc(2));

	    (*vz)[0] = static_cast<typename W::basicType>(vx->val());
	    (*vz)[1] = static_cast<typename W::basicType>(vy->val());
	    return vz;
	}
    };

    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct ConcatenateVectorVectorFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;
	    RCPtr<W> vz(W::alloc(vx->size() + vy->size()));

	    int j = 0;
	    for (int i = 0; i < vx->size(); i++, j++) {
		(*vz)[j] = static_cast<typename Z::basicType>((*vx)[i]);
	    }
	    for (int i = 0; i < vy->size(); i++, j++) {
		(*vz)[j] = static_cast<typename Z::basicType>((*vy)[i]);
	    }
	    return vz;
	}
    };

    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct ConcatenateVectorScalarFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;
	    RCPtr<W> vz(W::alloc(vx->size() + 1));

	    int j = 0;
	    for (int i = 0; i < vx->size(); i++, j++) {
		(*vz)[j] = static_cast<typename Z::basicType>((*vx)[i]);
	    }
	    (*vz)[j] = static_cast<typename Z::basicType>(vy->val());
	    return vz;
	}
    };

    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct ConcatenateScalarVectorFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;
	    RCPtr<W> vz(Z::alloc(1 + vy->size()));

	    int j = 0;
	    (*vz)[j++] = static_cast<typename Z::basicType>(vx->val());
	    for (int i = 0; i < vy->size(); i++, j++) {
		(*vz)[j] = static_cast<typename Z::basicType>((*vy)[i]);
	    }
	    return vz;
	}
    };
    
    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct ConcatenateObjectVectorsFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;
	    RCPtr<W> vz(W::alloc(vx->size() + vy->size()));

	    int j = 0;
	    for (int i = 0; i < vx->size(); i++, j++) {
		(*vz)[j] = static_cast<typename Z::basicType>((*vx)[i]->clone());
	    }
	    for (int i = 0; i < vy->size(); i++, j++) {
		(*vz)[j] = static_cast<typename Z::basicType>((*vy)[i]->clone());
	    }
	    return vz;
	}
    };

    template<class X, class Y, class Z, class W,
	     template<typename, typename, typename> typename OP = NullOperator>
    struct ConcatenateStringsFunctor
    {
	static ObjectRef Function(ObjectRef x, ObjectRef y)
	{
	    RCPtr<X> vx = x;
	    RCPtr<Y> vy = y;
	    RCPtr<W> vz(new String(*vx + *vy));
	    return vz;
	}
    };

    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_scalar_scalar_real(FunctorDispatch &dispatcher)
    {
	dispatcher.addFunctor<SI, SI, SI, SI, F, OP>();
	dispatcher.addFunctor<SI, SF, SF, SF, F, OP>();
	dispatcher.addFunctor<SI, SD, SD, SD, F, OP>();
	dispatcher.addFunctor<SF, SI, SF, SF, F, OP>();
	dispatcher.addFunctor<SF, SF, SF, SF, F, OP>();
	dispatcher.addFunctor<SF, SD, SD, SD, F, OP>();
	dispatcher.addFunctor<SD, SI, SD, SD, F, OP>();
	dispatcher.addFunctor<SD, SF, SD, SD, F, OP>();
	dispatcher.addFunctor<SD, SD, SD, SD, F, OP>();
    }

    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_scalar_scalar_bool(FunctorDispatch &dispatcher)
    {
	dispatcher.addFunctor<SI, SI, SI, B, F, OP>();
	dispatcher.addFunctor<SI, SF, SF, B, F, OP>();
	dispatcher.addFunctor<SI, SD, SD, B, F, OP>();
	dispatcher.addFunctor<SF, SI, SF, B, F, OP>();
	dispatcher.addFunctor<SF, SF, SF, B, F, OP>();
	dispatcher.addFunctor<SF, SD, SD, B, F, OP>();
	dispatcher.addFunctor<SD, SI, SD, B, F, OP>();
	dispatcher.addFunctor<SD, SF, SD, B, F, OP>();
	dispatcher.addFunctor<SD, SD, SD, B, F, OP>();
    }

    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_scalar_scalar(FunctorDispatch &dispatcher)
    {
	registerFunctor_scalar_scalar_real<F, OP>(dispatcher);

	dispatcher.addFunctor<SI, SG, SG, SG, F, OP>();
	dispatcher.addFunctor<SI, SE, SE, SE, F, OP>();
	dispatcher.addFunctor<SF, SG, SG, SG, F, OP>();
	dispatcher.addFunctor<SF, SE, SE, SE, F, OP>();
	dispatcher.addFunctor<SD, SG, SE, SE, F, OP>();
	dispatcher.addFunctor<SD, SE, SE, SE, F, OP>();

	dispatcher.addFunctor<SG, SI, SG, SG, F, OP>();
	dispatcher.addFunctor<SE, SI, SE, SE, F, OP>();
	dispatcher.addFunctor<SG, SF, SG, SG, F, OP>();
	dispatcher.addFunctor<SE, SF, SE, SE, F, OP>();
	dispatcher.addFunctor<SG, SD, SE, SE, F, OP>();
	dispatcher.addFunctor<SE, SD, SE, SE, F, OP>();
	
	dispatcher.addFunctor<SG, SG, SG, SG, F, OP>();
	dispatcher.addFunctor<SE, SG, SE, SE, F, OP>();
	dispatcher.addFunctor<SG, SE, SE, SE, F, OP>();
	dispatcher.addFunctor<SE, SE, SE, SE, F, OP>();
    }
    
    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_scalar_scalar_vector(FunctorDispatch &dispatcher)
    {
	dispatcher.addFunctor<SI, SI, VI, VI, F, OP>();
	dispatcher.addFunctor<SI, SF, VF, VF, F, OP>();
	dispatcher.addFunctor<SI, SD, VD, VD, F, OP>();
	dispatcher.addFunctor<SF, SI, VF, VF, F, OP>();
	dispatcher.addFunctor<SF, SF, VF, VF, F, OP>();
	dispatcher.addFunctor<SF, SD, VD, VD, F, OP>();
	dispatcher.addFunctor<SD, SI, VD, VD, F, OP>();
	dispatcher.addFunctor<SD, SF, VD, VD, F, OP>();
	dispatcher.addFunctor<SD, SD, VD, VD, F, OP>();

	dispatcher.addFunctor<SI, SG, VG, VG, F, OP>();
	dispatcher.addFunctor<SI, SE, VE, VE, F, OP>();
	dispatcher.addFunctor<SF, SG, VG, VG, F, OP>();
	dispatcher.addFunctor<SF, SE, VE, VE, F, OP>();
	dispatcher.addFunctor<SD, SG, VE, VE, F, OP>();
	dispatcher.addFunctor<SD, SE, VE, VE, F, OP>();
	
	dispatcher.addFunctor<SG, SI, VG, VG, F, OP>();
	dispatcher.addFunctor<SE, SI, VE, VE, F, OP>();
	dispatcher.addFunctor<SG, SF, VG, VG, F, OP>();
	dispatcher.addFunctor<SE, SF, VE, VE, F, OP>();
	dispatcher.addFunctor<SG, SD, VE, VE, F, OP>();
	dispatcher.addFunctor<SE, SD, VE, VE, F, OP>();

	dispatcher.addFunctor<SG, SG, VG, VG, F, OP>();
	dispatcher.addFunctor<SE, SG, VE, VE, F, OP>();
	dispatcher.addFunctor<SG, SE, VE, VE, F, OP>();
	dispatcher.addFunctor<SE, SE, VE, VE, F, OP>();
    }
    
    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_vector_vector_real(FunctorDispatch &dispatcher)
    {
	dispatcher.addFunctor<VI, VI, VI, VI, F, OP>();
	dispatcher.addFunctor<VI, VF, VF, VF, F, OP>();
	dispatcher.addFunctor<VI, VD, VD, VD, F, OP>();
	dispatcher.addFunctor<VF, VI, VF, VF, F, OP>();
	dispatcher.addFunctor<VF, VF, VF, VF, F, OP>();
	dispatcher.addFunctor<VF, VD, VD, VD, F, OP>();
	dispatcher.addFunctor<VD, VI, VD, VD, F, OP>();
	dispatcher.addFunctor<VD, VF, VD, VD, F, OP>();
	dispatcher.addFunctor<VD, VD, VD, VD, F, OP>();
    }

    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_vector_vector(FunctorDispatch &dispatcher)
    {
	registerFunctor_vector_vector_real<F, OP>(dispatcher);

	dispatcher.addFunctor<VI, VG, VG, VG, F, OP>();
	dispatcher.addFunctor<VI, VE, VE, VE, F, OP>();
	dispatcher.addFunctor<VF, VG, VG, VG, F, OP>();
	dispatcher.addFunctor<VF, VE, VE, VE, F, OP>();
	dispatcher.addFunctor<VD, VG, VE, VE, F, OP>();
	dispatcher.addFunctor<VD, VE, VE, VE, F, OP>();

	dispatcher.addFunctor<VG, VI, VG, VG, F, OP>();
	dispatcher.addFunctor<VE, VI, VE, VE, F, OP>();
	dispatcher.addFunctor<VG, VF, VG, VG, F, OP>();
	dispatcher.addFunctor<VE, VF, VE, VE, F, OP>();
	dispatcher.addFunctor<VG, VD, VE, VE, F, OP>();
	dispatcher.addFunctor<VE, VD, VE, VE, F, OP>();

	dispatcher.addFunctor<VG, VG, VG, VG, F, OP>();
	dispatcher.addFunctor<VE, VG, VE, VE, F, OP>();
	dispatcher.addFunctor<VG, VE, VE, VE, F, OP>();
	dispatcher.addFunctor<VE, VE, VE, VE, F, OP>();
    }
    
    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_scalar_vector(FunctorDispatch &dispatcher)
    {
	dispatcher.addFunctor<SI, VI, VI, VI, F, OP>();
	dispatcher.addFunctor<SI, VF, VF, VF, F, OP>();
	dispatcher.addFunctor<SI, VD, VD, VD, F, OP>();
	dispatcher.addFunctor<SF, VI, VF, VF, F, OP>();
	dispatcher.addFunctor<SF, VF, VF, VF, F, OP>();
	dispatcher.addFunctor<SF, VD, VD, VD, F, OP>();
	dispatcher.addFunctor<SD, VI, VD, VD, F, OP>();
	dispatcher.addFunctor<SD, VF, VD, VD, F, OP>();
	dispatcher.addFunctor<SD, VD, VD, VD, F, OP>();

	dispatcher.addFunctor<SI, VG, VG, VG, F, OP>();
	dispatcher.addFunctor<SI, VE, VE, VE, F, OP>();
	dispatcher.addFunctor<SF, VG, VG, VG, F, OP>();
	dispatcher.addFunctor<SF, VE, VE, VE, F, OP>();
	dispatcher.addFunctor<SD, VG, VE, VE, F, OP>();
	dispatcher.addFunctor<SD, VE, VE, VE, F, OP>();
	
	dispatcher.addFunctor<SG, VI, VG, VG, F, OP>();
	dispatcher.addFunctor<SE, VI, VE, VE, F, OP>();
	dispatcher.addFunctor<SG, VF, VG, VG, F, OP>();
	dispatcher.addFunctor<SE, VF, VE, VE, F, OP>();
	dispatcher.addFunctor<SG, VD, VE, VE, F, OP>();
	dispatcher.addFunctor<SE, VD, VE, VE, F, OP>();

	dispatcher.addFunctor<SG, VG, VG, VG, F, OP>();
	dispatcher.addFunctor<SE, VG, VE, VE, F, OP>();
	dispatcher.addFunctor<SG, VE, VE, VE, F, OP>();
	dispatcher.addFunctor<SE, VE, VE, VE, F, OP>();
    }
    
    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_vector_scalar(FunctorDispatch &dispatcher)
    {
	dispatcher.addFunctor<VI, SI, VI, VI, F, OP>();
	dispatcher.addFunctor<VI, SF, VF, VF, F, OP>();
	dispatcher.addFunctor<VI, SD, VD, VD, F, OP>();
	dispatcher.addFunctor<VF, SI, VF, VF, F, OP>();
	dispatcher.addFunctor<VF, SF, VF, VF, F, OP>();
	dispatcher.addFunctor<VF, SD, VD, VD, F, OP>();
	dispatcher.addFunctor<VD, SI, VD, VD, F, OP>();
	dispatcher.addFunctor<VD, SF, VD, VD, F, OP>();
	dispatcher.addFunctor<VD, SD, VD, VD, F, OP>();

	dispatcher.addFunctor<VI, SG, VG, VG, F, OP>();
	dispatcher.addFunctor<VI, SE, VE, VE, F, OP>();
	dispatcher.addFunctor<VF, SG, VG, VG, F, OP>();
	dispatcher.addFunctor<VF, SE, VE, VE, F, OP>();
	dispatcher.addFunctor<VD, SG, VE, VE, F, OP>();
	dispatcher.addFunctor<VD, SE, VE, VE, F, OP>();
	
	dispatcher.addFunctor<VG, SI, VG, VG, F, OP>();
	dispatcher.addFunctor<VE, SI, VE, VE, F, OP>();
	dispatcher.addFunctor<VG, SF, VG, VG, F, OP>();
	dispatcher.addFunctor<VE, SF, VE, VE, F, OP>();
	dispatcher.addFunctor<VG, SD, VE, VE, F, OP>();
	dispatcher.addFunctor<VE, SD, VE, VE, F, OP>();

	dispatcher.addFunctor<VG, SG, VG, VG, F, OP>();
	dispatcher.addFunctor<VE, SG, VE, VE, F, OP>();
	dispatcher.addFunctor<VG, SE, VE, VE, F, OP>();
	dispatcher.addFunctor<VE, SE, VE, VE, F, OP>();
    }
    
    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_matrix_matrix_real(FunctorDispatch &dispatcher)
    {
	dispatcher.addFunctor<MI, MI, MI, MI, F, OP>();
	dispatcher.addFunctor<MI, MF, MF, MF, F, OP>();
	dispatcher.addFunctor<MI, MD, MD, MD, F, OP>();
	dispatcher.addFunctor<MF, MI, MF, MF, F, OP>();
	dispatcher.addFunctor<MF, MF, MF, MF, F, OP>();
	dispatcher.addFunctor<MF, MD, MD, MD, F, OP>();
	dispatcher.addFunctor<MD, MI, MD, MD, F, OP>();
	dispatcher.addFunctor<MD, MF, MD, MD, F, OP>();
	dispatcher.addFunctor<MD, MD, MD, MD, F, OP>();
    }

    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_matrix_matrix(FunctorDispatch &dispatcher)
    {
	registerFunctor_matrix_matrix_real<F, OP>(dispatcher);

	dispatcher.addFunctor<MI, MG, MG, MG, F, OP>();
	dispatcher.addFunctor<MI, ME, ME, ME, F, OP>();
	dispatcher.addFunctor<MF, MG, MG, MG, F, OP>();
	dispatcher.addFunctor<MF, ME, ME, ME, F, OP>();
	dispatcher.addFunctor<MD, MG, ME, ME, F, OP>();
	dispatcher.addFunctor<MD, ME, ME, ME, F, OP>();

	dispatcher.addFunctor<MG, MI, MG, MG, F, OP>();
	dispatcher.addFunctor<ME, MI, ME, ME, F, OP>();
	dispatcher.addFunctor<MG, MF, MG, MG, F, OP>();
	dispatcher.addFunctor<ME, MF, ME, ME, F, OP>();
	dispatcher.addFunctor<MG, MD, ME, ME, F, OP>();
	dispatcher.addFunctor<ME, MD, ME, ME, F, OP>();

	dispatcher.addFunctor<MG, MG, MG, MG, F, OP>();
	dispatcher.addFunctor<ME, MG, ME, ME, F, OP>();
	dispatcher.addFunctor<MG, ME, ME, ME, F, OP>();
	dispatcher.addFunctor<ME, ME, ME, ME, F, OP>();
    }
    
    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_scalar_matrix(FunctorDispatch &dispatcher)
    {
	dispatcher.addFunctor<SI, MI, MI, MI, F, OP>();
	dispatcher.addFunctor<SI, MF, MF, MF, F, OP>();
	dispatcher.addFunctor<SI, MD, MD, MD, F, OP>();
	dispatcher.addFunctor<SF, MI, MF, MF, F, OP>();
	dispatcher.addFunctor<SF, MF, MF, MF, F, OP>();
	dispatcher.addFunctor<SF, MD, MD, MD, F, OP>();
	dispatcher.addFunctor<SD, MI, MD, MD, F, OP>();
	dispatcher.addFunctor<SD, MF, MD, MD, F, OP>();
	dispatcher.addFunctor<SD, MD, MD, MD, F, OP>();

	dispatcher.addFunctor<SI, MG, MG, MG, F, OP>();
	dispatcher.addFunctor<SI, ME, ME, ME, F, OP>();
	dispatcher.addFunctor<SF, MG, MG, MG, F, OP>();
	dispatcher.addFunctor<SF, ME, ME, ME, F, OP>();
	dispatcher.addFunctor<SD, MG, ME, ME, F, OP>();
	dispatcher.addFunctor<SD, ME, ME, ME, F, OP>();
	
	dispatcher.addFunctor<SG, MI, MG, MG, F, OP>();
	dispatcher.addFunctor<SE, MI, ME, ME, F, OP>();
	dispatcher.addFunctor<SG, MF, MG, MG, F, OP>();
	dispatcher.addFunctor<SE, MF, ME, ME, F, OP>();
	dispatcher.addFunctor<SG, MD, ME, ME, F, OP>();
	dispatcher.addFunctor<SE, MD, ME, ME, F, OP>();

	dispatcher.addFunctor<SG, MG, MG, MG, F, OP>();
	dispatcher.addFunctor<SE, MG, ME, ME, F, OP>();
	dispatcher.addFunctor<SG, ME, ME, ME, F, OP>();
	dispatcher.addFunctor<SE, ME, ME, ME, F, OP>();
    }
    
    template<template<typename, typename, typename, typename, template<typename, typename, typename> typename> typename F,
	     template<typename, typename, typename> typename OP = NullOperator>
    static void registerFunctor_matrix_scalar(FunctorDispatch &dispatcher)
    {
	dispatcher.addFunctor<MI, SI, MI, MI, F, OP>();
	dispatcher.addFunctor<MI, SF, MF, MF, F, OP>();
	dispatcher.addFunctor<MI, SD, MD, MD, F, OP>();
	dispatcher.addFunctor<MF, SI, MF, MF, F, OP>();
	dispatcher.addFunctor<MF, SF, MF, MF, F, OP>();
	dispatcher.addFunctor<MF, SD, MD, MD, F, OP>();
	dispatcher.addFunctor<MD, SI, MD, MD, F, OP>();
	dispatcher.addFunctor<MD, SF, MD, MD, F, OP>();
	dispatcher.addFunctor<MD, SD, MD, MD, F, OP>();

	dispatcher.addFunctor<MI, SG, MG, MG, F, OP>();
	dispatcher.addFunctor<MI, SE, ME, ME, F, OP>();
	dispatcher.addFunctor<MF, SG, MG, MG, F, OP>();
	dispatcher.addFunctor<MF, SE, ME, ME, F, OP>();
	dispatcher.addFunctor<MD, SG, ME, ME, F, OP>();
	dispatcher.addFunctor<MD, SE, ME, ME, F, OP>();
	
	dispatcher.addFunctor<MG, SI, MG, MG, F, OP>();
	dispatcher.addFunctor<ME, SI, ME, ME, F, OP>();
	dispatcher.addFunctor<MG, SF, MG, MG, F, OP>();
	dispatcher.addFunctor<ME, SF, ME, ME, F, OP>();
	dispatcher.addFunctor<MG, SD, ME, ME, F, OP>();
	dispatcher.addFunctor<ME, SD, ME, ME, F, OP>();

	dispatcher.addFunctor<MG, SG, MG, MG, F, OP>();
	dispatcher.addFunctor<ME, SG, ME, ME, F, OP>();
	dispatcher.addFunctor<MG, SE, ME, ME, F, OP>();
	dispatcher.addFunctor<ME, SE, ME, ME, F, OP>();
    }
    
    class AddFunctorDispatch : public FunctorDispatch
    {
    public:
	AddFunctorDispatch(){
	    registerFunctor_scalar_scalar<ScalarScalarFunctor, AddOperator>(*this);
	    registerFunctor_vector_vector<VectorVectorFunctor, AddOperator>(*this);
	    registerFunctor_scalar_vector<ScalarVectorFunctor, AddOperator>(*this);
	    registerFunctor_vector_scalar<VectorScalarFunctor, AddOperator>(*this);
	    registerFunctor_matrix_matrix<MatrixMatrixFunctor, AddOperator>(*this);
	    registerFunctor_scalar_matrix<ScalarMatrixFunctor, AddOperator>(*this);
	    registerFunctor_matrix_scalar<MatrixScalarFunctor, AddOperator>(*this);
	}
    };

    class SubFunctorDispatch : public FunctorDispatch
    {
    public:
	SubFunctorDispatch(){
	    registerFunctor_scalar_scalar<ScalarScalarFunctor, SubOperator>(*this);
	    registerFunctor_vector_vector<VectorVectorFunctor, SubOperator>(*this);
	    registerFunctor_scalar_vector<ScalarVectorFunctor, SubOperator>(*this);
	    registerFunctor_vector_scalar<VectorScalarFunctor, SubOperator>(*this);
	    registerFunctor_matrix_matrix<MatrixMatrixFunctor, SubOperator>(*this);
	    registerFunctor_scalar_matrix<ScalarMatrixFunctor, SubOperator>(*this);
	    registerFunctor_matrix_scalar<MatrixScalarFunctor, SubOperator>(*this);
	}
    };
       
    class MulFunctorDispatch : public FunctorDispatch
    {
    public:
	MulFunctorDispatch(){
	    registerFunctor_scalar_scalar<ScalarScalarFunctor, MulOperator>(*this);
	    registerFunctor_vector_vector<VectorVectorFunctor, MulOperator>(*this);
	    registerFunctor_scalar_vector<ScalarVectorFunctor, MulOperator>(*this);
	    registerFunctor_vector_scalar<VectorScalarFunctor, MulOperator>(*this);
	    registerFunctor_matrix_matrix<MatrixMatrixFunctor, MulOperator>(*this);
	    registerFunctor_scalar_matrix<ScalarMatrixFunctor, MulOperator>(*this);
	    registerFunctor_matrix_scalar<MatrixScalarFunctor, MulOperator>(*this);
	}
    };

    class DivFunctorDispatch : public FunctorDispatch
    {
    public:
	DivFunctorDispatch(){
	    registerFunctor_scalar_scalar<ScalarScalarFunctor, DivOperator>(*this);
	    registerFunctor_vector_vector<VectorVectorFunctor, DivOperator>(*this);
	    registerFunctor_scalar_vector<ScalarVectorFunctor, DivOperator>(*this);
	    registerFunctor_vector_scalar<VectorScalarFunctor, DivOperator>(*this);
	    registerFunctor_matrix_matrix<MatrixMatrixFunctor, DivOperator>(*this);
	    registerFunctor_scalar_matrix<ScalarMatrixFunctor, DivOperator>(*this);
	    registerFunctor_matrix_scalar<MatrixScalarFunctor, DivOperator>(*this);
	}
    };
       
    class MinFunctorDispatch : public FunctorDispatch
    {
    public:
	MinFunctorDispatch(){
	    registerFunctor_scalar_scalar_real<ScalarScalarFunctor, MinOperator>(*this);
	    registerFunctor_vector_vector_real<VectorVectorFunctor, MinOperator>(*this);
	    registerFunctor_matrix_matrix_real<MatrixMatrixFunctor, MinOperator>(*this);
	}
    };

    class MaxFunctorDispatch : public FunctorDispatch
    {
    public:
	MaxFunctorDispatch(){
	    registerFunctor_scalar_scalar_real<ScalarScalarFunctor, MaxOperator>(*this);
	    registerFunctor_vector_vector_real<VectorVectorFunctor, MaxOperator>(*this);
	    registerFunctor_matrix_matrix_real<MatrixMatrixFunctor, MaxOperator>(*this);
	}
    };
       
    class LessFunctorDispatch : public FunctorDispatch
    {
    public:
	LessFunctorDispatch(){
	    registerFunctor_scalar_scalar_bool<ScalarScalarFunctor, LessOperator>(*this);
	}
    };

    class GreaterFunctorDispatch : public FunctorDispatch
    {
    public:
	GreaterFunctorDispatch(){
	    registerFunctor_scalar_scalar_bool<ScalarScalarFunctor, GreaterOperator>(*this);
	}
    };
       
    class EqualFunctorDispatch : public FunctorDispatch
    {
    public:
	EqualFunctorDispatch(){
	    registerFunctor_scalar_scalar_bool<ScalarScalarFunctor, EqualOperator>(*this);
	}
    };
       
    class ConcatenateFunctorDispatch : public FunctorDispatch
    {
    public:
	ConcatenateFunctorDispatch(){
	    registerFunctor_scalar_scalar_vector<ConcatenateScalarScalarFunctor>(*this);
	    registerFunctor_scalar_vector<ConcatenateScalarVectorFunctor>(*this);
	    registerFunctor_vector_scalar<ConcatenateVectorScalarFunctor>(*this);
	    registerFunctor_vector_vector<ConcatenateVectorVectorFunctor>(*this);

	    this->addFunctor<Vector<ObjectRef>, Vector<ObjectRef>, Vector<ObjectRef>, Vector<ObjectRef>, ConcatenateObjectVectorsFunctor, NullOperator>();
	    this->addFunctor<String, String, String, String, ConcatenateStringsFunctor, NullOperator>();
	}
    };
       
    static OperatorDispatch theOperatorDispatch;
    static EqualFunctorDispatch theEqualFunctorDispatch;
    static AddFunctorDispatch theAddFunctorDispatch;
    static SubFunctorDispatch theSubFunctorDispatch;
    static MulFunctorDispatch theMulFunctorDispatch;
    static DivFunctorDispatch theDivFunctorDispatch;
    static LessFunctorDispatch theLessFunctorDispatch;
    static GreaterFunctorDispatch theGreaterFunctorDispatch;
    static MinFunctorDispatch theMinFunctorDispatch;
    static MaxFunctorDispatch theMaxFunctorDispatch;
    static ConcatenateFunctorDispatch theConcatenateFunctorDispatch;

    OperatorDispatch::OperatorDispatch()
    {
	this->dispatchTable[EQUAL]   = &theEqualFunctorDispatch;
	this->dispatchTable[ADD]     = &theAddFunctorDispatch;
	this->dispatchTable[SUB]     = &theSubFunctorDispatch;
	this->dispatchTable[MUL]     = &theMulFunctorDispatch;
	this->dispatchTable[DIV]     = &theDivFunctorDispatch;
	this->dispatchTable[LESS]    = &theLessFunctorDispatch;
	this->dispatchTable[GREATER] = &theGreaterFunctorDispatch;
	this->dispatchTable[MIN]     = &theMinFunctorDispatch;
	this->dispatchTable[MAX]     = &theMaxFunctorDispatch;
	this->dispatchTable[CONCAT]  = &theConcatenateFunctorDispatch;
    }

    ObjectRef
    operator + (ObjectRef x, ObjectRef y)
    {
	return theOperatorDispatch[OperatorDispatch::ADD].call(x, y);
    }

    ObjectRef
    operator - (ObjectRef x, ObjectRef y)
    {
	return theOperatorDispatch[OperatorDispatch::SUB].call(x, y);
    }

    ObjectRef
    operator * (ObjectRef x, ObjectRef y)
    {
	return theOperatorDispatch[OperatorDispatch::MUL].call(x, y);
    }

    ObjectRef
    operator / (ObjectRef x, ObjectRef y)
    {
	return theOperatorDispatch[OperatorDispatch::DIV].call(x, y);
    }

    ObjectRef
    operator < (ObjectRef x, ObjectRef y)
    {
	return theOperatorDispatch[OperatorDispatch::LESS].call(x, y);
    }

    ObjectRef
    operator > (ObjectRef x, ObjectRef y)
    {
	return theOperatorDispatch[OperatorDispatch::GREATER].call(x, y);
    }

    ObjectRef
    operator == (ObjectRef x, ObjectRef y)
    {
	return theOperatorDispatch[OperatorDispatch::EQUAL].call(x, y);
    }

    ObjectRef
    max(ObjectRef x, ObjectRef y)
    {
	return theOperatorDispatch[OperatorDispatch::MAX].call(x, y);
    }

    ObjectRef
    min(ObjectRef x, ObjectRef y)
    {
	return theOperatorDispatch[OperatorDispatch::MIN].call(x, y);
    }

    ObjectRef
    concat(ObjectRef x, ObjectRef y)
    {
	return theOperatorDispatch[OperatorDispatch::CONCAT].call(x, y);
    }
}

/* end of file */
