#!/usr/bin/python3 -u
#
# Copyright (c) 2008 - 2020 Kyoto University and Honda Motor Co.,Ltd.
# All rights reserved.
# HARK was developed by researchers in Okuno Laboratory
# at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
# 
# Licensed under the HARK license, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
# https://www.hark.jp/notice/HARK_License_Agreement.pdf
#
# Parses javadoc-like comments from C++ files and writes them into well-formed
# (hopefully) XML NodeClass definitions, that can then be used by Overflow to
# extract info about the nodes.
#
# Usage:
# Takes a file as input and prints out the XML, so the read the info from
# to_read.cc and output it to nice_xml.def you would need to do something like:
# 'python info2def.py to_read.cc > nice_xml.def.'

from __future__ import print_function

import sys
import re
import fileinput
import copy

print("<?xml version=\"1.0\"?>")
print("<Definitions>")

# data structures to store the info for a node. They look like the following:
#
# info{name} -> name
# info{category} -> category
# info{description} -> description
# info{inputs} -> list of input info
#      info{inputs}[n]{input_name} -> name
#      info{inputs}[n]{input_type} -> type
#      info{inputs}[n]{input_description} -> description
# info{outputs} -> list of output info
#      info{outputs}[n]{output_name} -> name
#      info{outputs}[n]{output_type} -> type
#      info{outputs}[n]{output_description} -> description
# info{params} -> list of parameter info
#      info{params}[n]{parameter_name} -> name
#      info{params}[n]{parameter_type} -> type
#      info{params}[n]{parameter_value} -> value
#      info{params}[n]{parameter_list} -> list
#      info{params}[n]{parameter_valid} -> valid
#      info{params}[n]{parameter_description} -> description

default_input = {
    'input_name': "No Name",
    'input_type': "any",
    'input_description': "No Description Available"
}

default_output = {
    'output_name': "No Name",
    'output_type': "any",
    'output_description': "No Description Available"
}

default_param = {
    'parameter_name': "No Name",
    'parameter_type': "any",
    'parameter_value': "",
    'parameter_list': "",
    'parameter_valid': "",
    'parameter_description': "No Description Available"
}

default_node_info = {
    'name': "No Name",
    'category': "No Category",
    'description': "No Description Available",
    'inputs': [],
    'outputs': [],
    'params': []
}
node_info = copy.deepcopy(default_node_info)

# the last info parsed, so we can deal with carryovers.
last_info = ""
last_name = ""

# flag so we know if we are in a javadoc comment
in_comment = 0

# assign the inputs and outputs to the default value
current_input  = copy.deepcopy(default_input)
current_output = copy.deepcopy(default_output)
current_param  = copy.deepcopy(default_param)

dep = {}
filedep = {}
headerdep = {}

# read through the file
for line in fileinput.input():
    base_filename = fileinput.filename()

    # if we are in a comment, look for the information to parse.
    if in_comment == 1 :

        # look for a '* @something info', then we need to process the item.
        m = re.search("\* (@[a-zA-Z:_0-9]+) +([\<\>a-zA-Z:_0-9 .,;\"\'\/\*\!\+=\-\(\)\[\]\{\}]+)", line)
        if m is not None :
            pname = m.group(1)
            pval  = m.group(2)
            pval = re.sub("\s*$", "", pval, 1)
            pval = re.sub("<", "&lt;", pval)
            pval = re.sub(">", "&gt;", pval)

            if pname == '@name' :
                node_info[pname[1:]] = pval
            elif pname == '@category' :
                node_info[pname[1:]]= pval
            elif pname == '@require' :
                if base_filename in filedep :
                    filedep[base_filename].append(pval)
                else :
                    filedep[base_filename] = [pval]
                node_info[pname[1:]] = pval
            elif pname == '@description' :
                node_info[pname[1:]] = pval

            # start of a new input
            elif pname == '@input_name' :
                # check if we have an old input we need to append
                if current_input['input_name'] != 'No Name' :
                    # add the current input to the inputs
                    node_info['inputs'].append(copy.deepcopy(current_input))
                    # clear the input
                    current_input = copy.deepcopy(default_input)
                current_input[pname[1:]] = pval
            elif pname == '@input_type' :
                current_input[pname[1:]] = pval
            elif pname == '@input_description' :
                current_input[pname[1:]] = pval

            # start of a new output
            elif pname == '@output_name' :
                # check if we have an old output we need to append
                if current_output['output_name'] != 'No Name' :
                    # add the current output to the outputs
                    node_info['outputs'].append(copy.deepcopy(current_output))
                    # clear the info
                    current_output = copy.deepcopy(default_output)
                current_output[pname[1:]] = pval

            elif pname == '@output_type' :
                current_output[pname[1:]] = pval

            elif  pname == '@output_description' :
                current_output[pname[1:]] = pval

            # start of a new parameter
            elif  pname == '@parameter_name' :
                # check if we have an old parameter to append to the info
                if  current_param['parameter_name'] != 'No Name' :
                    # add the current parameter to the parameter list
                    node_info['params'].append(copy.deepcopy(current_param))
                    # clear the parameter info
                    current_param = copy.deepcopy(default_param)
                current_param[pname[1:]] = pval
            elif pname == '@parameter_type' :
                current_param[pname[1:]] = pval
            elif  pname == '@parameter_value' :
                current_param[pname[1:]] = pval
            elif pname == '@parameter_description' :
                current_param[pname[1:]] = pval
            elif pname == '@parameter_list' :
                current_param[pname[1:]] = pval
            elif  pname == '@parameter_valid' :
                current_param[pname[1:]] = pval
            else:
                print("Unrecognized \"@\" tag: {}".format(pname), file=sys.stderr)
                sys.exit()

            # set the added information as the last info added
            last_name = pname[1:]
            last_info = pval

        elif re.search("END\*\/", line) is not None : # look for the trailing '*/' 
            if current_input['input_name'] != 'No Name' :
                node_info['inputs'].append(copy.deepcopy(current_input))
            if current_output['output_name'] != 'No Name' :
                node_info['outputs'].append(copy.deepcopy(current_output))
            if current_param['parameter_name'] != 'No Name' :
                node_info['params'].append(copy.deepcopy(current_param))

            print("  <NodeClass name=\"{}\" category=\"{}\"".format(node_info['name'], node_info['category']), end="")
            print(" source=\"{}\"".format(base_filename), end="")
            if 'require' in node_info :
                print(" require=\"{}\"".format(node_info['require']), end="")
            print(">")

            for _input in node_info['inputs'] :
                print("    <Input name=\"{}\" type=\"{}\">".format(_input['input_name'], _input['input_type']), end="")
                print(_input['input_description'], end="")
                print("</Input>")

            for _output in node_info['outputs'] :
                print("    <Output name=\"{}\" type=\"{}\">".format(_output['output_name'], _output['output_type']), end="")
                print(_output['output_description'], end="")
                print("</Output>")

            for param in node_info['params'] :
                print("    <Parameter name=\"{}\" type=\"{}\" value=\"{}\" list=\"{}\" valid=\"{}\">" \
                  .format(param['parameter_name'], param['parameter_type'], param['parameter_value'], \
                    param['parameter_list'], param['parameter_valid']), end="")
                print(param['parameter_description'], end="")
                print("</Parameter>")

            print("    <Description>", end="")
            print(node_info['description'], end="")
            print("</Description>")
            print("  </NodeClass>")

            # the last info parsed, so we can deal with carryovers.
            last_info = ""
            last_name = ""

            # flag so we know if we are in a javadoc comment
            in_comment = 0

            # assign the inputs and outputs to the default value
            current_input  = copy.deepcopy(default_input)
            current_output = copy.deepcopy(default_output)
            current_param  = copy.deepcopy(default_param)
            node_info      = copy.deepcopy(default_node_info)
            in_comment     = 0

        else :
            # line spill over from the previous comment. Need to add this
            # info to the previous value
            m = re.search("\* ([\<\>a-zA-Z:_0-9 .,\"\'\//\*\!\+=\-\(\)\[\]\{\}]+)", line)
            if m is not None :
                new_info = last_info + "\n" + m.group(1)
                # if it is input info add it to the current_input
                if last_name[0:5] == 'input' :
                    current_input[last_name] = new_info
                # if it is an output, add it to the current output
                elif last_name[0:6] == 'output' :
                    current_output[last_name] = new_info
                # if it is a param, add it to the current param
                elif last_name[0:9] == 'parameter' :
                    current_param[last_name] = new_info
                # otherwise, we can add it directly to the node info
                else :
                    node_info[last_name] = new_info
                # keep expanding last_info for multiple carryovers
                last_info = new_info
            #elif re.search("\*", line) :
                #print("blank line ")

    # if we aren't in a comment, then keep looking for the beginning
    elif re.search("\/\*Node", line) is not None :
        #print("comment\n")
        in_comment = 1
    else :
        m = re.search("\/\/\@implements ([a-zA-Z0-9_]*)", line)
        if m is not None :
            if m.group(1) in dep :
                dep[m.group(1)].append(base_filename)
            else :
                dep[m.group(1)] = [base_filename]
        else :
            m = re.search("\/\/\@require ([a-zA-Z0-9_]*)", line)
            if m is not None :
                if base_filename in filedep :
                    filedep[base_filename].append(m.group(1))
                else :
                    filedep[base_filename] = [m.group(1)]
            else :
                m = re.search("^\#include \"([a-zA-Z_0-9\.]*)\"", line)
                if m is not None :
                    if base_filename in headerdep :
                        headerdep[base_filename].append(m.group(1))
                    else :
                        headerdep[base_filename] = [m.group(1)]

#print("<Dependencies>")
for module in dep.keys() :
    print("  <ModuleDepend module=\"{}\">".format(module))
    for depend in dep[module] :
        print("    <Require file=\"{}\"/>".format(depend))
    print("  </ModuleDepend>")

for file in filedep.keys() :
    print("  <FileDepend file=\"{}\">".format(file))
    for module in filedep[file]:
        print("    <RequireModule module=\"{}\"/>".format(module))
    print("  </FileDepend>")

for file in headerdep.keys() :
    print("  <FileDepend file=\"{}\">".format(file))
    for header in headerdep[file] :
        print("    <RequireHeader header=\"{}\"/>".format(header))
    print("  </FileDepend>")

#print("</Dependencies>")

print("</Definitions>")
